# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines adapters for converting feature importance values to an explanation."""

from .explanation_adapter import ExplanationAdapter

__all__ = ["ExplanationAdapter"]
