# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the building blocks for explanations returned by explainers."""

from .serialization import load_explanation, save_explanation

__all__ = ['load_explanation', 'save_explanation']
