# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for permutation feature importance."""

from .permutation_importance import PFIExplainer

__all__ = ['PFIExplainer']
