# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for mimic explainer and explainable surrogate models."""
from .mimic_explainer import MimicExplainer

__all__ = ["MimicExplainer"]
