# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for Explanation Dashboard widget."""

from .explanation_dashboard import ExplanationDashboard

__all__ = ['ExplanationDashboard']
